#include "common.h"
#include <Rdefines.h>
#include <R_ext/Utils.h>
#include <fenv.h>
#include <db.h>

/**
 * Compute the log-sum of a bunch of numbers being stored as a log by subtracting out
 * the largest value and replacing it at the end of the calculations.
 *
 * @param v a (disposable) vector of values
 * @param n the length of the vector
 * @return log(exp(v[0]) + exp(v[1]) + ... + exp(v[n]))
 */
double logsumexp(double*v,int n) {
  double max=v[0],result = 0.0;
  int    i;
  if(n==1) return v[0];
  for(i=1;i<n;i++) if(!isnan(v[i]) && max < v[i]) max = v[i];
  if(max == R_NegInf) return R_NegInf;
  for(i=0;i<n;i++) {
    if(!isnan(v[i]) && v[i] != R_NegInf) result += exp(v[i]-max);
  }
  return log(result)+max;
}

double boltz_Z(double*v,double t,int n) {
  double max=v[0]/t,result = 0.0;
  int    i;
  if(n==1) return v[0]/t;
  for(i=1;i<n;i++) if(!isnan(v[i]) && max < v[i]/t) max = v[i]/t;
  for(i=0;i<n;i++) 
    if(!isnan(v[i])) result += exp((v[i]/t)-max);
  return log(result)+max;
}

double findmax(double*v,int n) {
  double max=v[0];
  int i;
  if(n==1) return max;
  for(i=1;i<n;i++) {
    if(v[i] > max) max = v[i];
  }
  return max;
}
double findmin(double*v,int n) {
  double min=v[0];
  int i;
  if(n==1) return min;
  for(i=1;i<n;i++)
    if(v[i] < min) min=v[i];
  return min;
}

SEXP BNSL_LogSumExp(SEXP v) {
  int    i,n=length(v);
  double max=REAL(v)[0],result=0.0;
  for(i=1;i<n;i++) if(REAL(v)[i] > max) max = REAL(v)[i];
  for(i=0;i<n;i++) result += exp(REAL(v)[i]-max);
  return ScalarReal(log(result)+max);
}

struct dd {
  unsigned int count;
  double       score;
};

/* Generate a unique set of objects (replaces BNSL_OrderSet) */
SEXP BNSL_UniqueSet(SEXP objs,SEXP score,SEXP _skip,SEXP _thin) {
  int eltsize;
  int type;
  type = TYPEOF(VECTOR_ELT(objs,0));
  switch(type) {
  case LGLSXP:case INTSXP:eltsize = sizeof(int);
  case REALSXP:eltsize= sizeof(double);
  case RAWSXP:eltsize = sizeof(char);
  default:
    error("Unsupported type for generating a unique set");
  }
  
  DB *db;DBT key,data;
  struct dd val;
  char *vmax = vmaxget();
  db_create(&db,NULL,0);
  db->open(db,NULL,NULL,NULL,DB_BTREE,DB_CREATE,0);
  int N    = length(objs);
  int skip = INTEGER(_skip)[0];
  int thin = INTEGER(_thin)[0];

  int i,unique = 0;
  memset(&key,0,sizeof(DBT));
  memset(&data,0,sizeof(DBT));
  for(i=skip;i<N;i+=thin) {
    SEXP obj = VECTOR_ELT(objs,i);
    key.data = RAW(obj);
    key.size = length(obj)*eltsize;
    if(db->get(db,NULL,&key,&data,0) == 0) {
      struct dd *tmp = (struct dd*)data.data;
      tmp->count++;
    } else {
      unique++;
      val.count = 1;
      if(score != R_NilValue) val.score = REAL(score)[i];
      data.data = &val;data.size = sizeof(struct dd);
    }
    db->put(db,NULL,&key,&data,0);
  }
  
  SEXP ret,count;
  PROTECT(ret = NEW_OBJECT(MAKE_CLASS(score == R_NilValue?"objectSet":"scoredObjectSet")));
  PROTECT(objs  = allocVector(VECSXP,unique));
  SET_SLOT(ret,install("object"),objs);
  PROTECT(count = allocVector(REALSXP,unique));
  SET_SLOT(ret,install("count"),count);
  if(score != R_NilValue) {
    PROTECT(score = allocVector(REALSXP,unique));
    SET_SLOT(ret,install("score"),score);	
    UNPROTECT(1);
  }
  DBC *cur;
  db->cursor(db,NULL,&cur,0);
  int  pos = 0;
  while(cur->c_get(cur,&key,&data,DB_NEXT) == 0) {
    struct dd* tmp = (struct dd*)data.data;
    SEXP obj;
    PROTECT(obj = allocVector(type,key.size/eltsize));
    memcpy(RAW(obj),key.data,key.size);
    SET_VECTOR_ELT(objs,i,obj);
    UNPROTECT(1);
    if(score != R_NilValue) REAL(score)[pos] = tmp->score;
    REAL(count)[pos++] = tmp->count;
  }
  cur->c_close(cur);
  vmaxset(vmax);
  db->close(db,0);
  UNPROTECT(3);
  return ret;
}
